;+
;Language : IDL 8.2.2
;================================================================================
;The AAS gives permission to anyone who wishes to use these subroutines to run their own calculations.
;Permission to republish or reuse these routines should be directed to permissions@aas.org.
;
;Note that the AAS does not take responsibility for the content of the source code. Potential users should
;be wary of applying the code to conditions that the code was not written to model and the accuracy of the
;code may be affected when compiled and executed on different systems.
;================================================================================
; NAME:
;       FIT_3D_ELLIPSOID
;
; PURPOSE:
;       Takes a set of cartesian data points as input, fits an free to rotate 3D ellipsoid, and returns the associated parameters.
;
; CALLING SEQUENCE:
;       fit_3D_ellipsoid, X, Y, Z, CENTER_XYZ, SIGMA_XYZ_P, rotmatrix[, euler_angles, XP, YP, ZP, EX=ex, EY=ey, EZ=ez, EXP=exp, EYP=eyp, EZP=ezp]
;
; INPUTS:
;       X, Y, Z = Set of cartesian coordinates. Each of them must be an array of the same size.
;
; OPTIONAL INPUT KEYWORDS:
;       EX, EY, EZ - Uncertainties on the X, Y, Z coordinates. *They will not be considered in the ellipsoid fitting*. However, they will be carried
;                    to the errors on the XP, YP, ZP coordinates in the new frame of reference.
;
; OUTPUTS:
;       CENTER_XYZ - A 3-elements array containing the center in the X, Y and Z directions, respectively. This is the only point where the X, Y, Z and
;                    XP, YP, ZP frames of reference match together.
;       SIGMA_XYZ_P - A 3-elements array containing the 1-sigma scatter in the XP, YP and ZP direction (i.e. in the rotated frame of reference). Note 
;                     that the direction with the largest scatter in the X,Y,Z frame will not necessarily match that of the XP,YP,ZP frame (e.g. if X
;                     is the direction with the largest scatter, it might be YP or ZP in the rotated frame).
;       ROTMATRIX - The 3 x 3 rotation matrix that allows the passage from the X,Y,Z frame of reference to the XP, YP, ZP frame, for which the axes point
;                   in the same directions than the eigenvectors of the X, Y, Z coordinates distribution.
;       EULER_ANGLES - A 3-elements array containing respectively the psi, theta and phi Euler angles (also called alpha, beta, gamma).
;       XP, YP, ZP - The positions of the X, Y, Z coordinates in the rotated frame of reference.
;       EXP, EYP, EZP - The EX, EY, EZ errors on coordinates, propagated in the rotated frame of reference.
;
; NOTE:
;
;       Please acknowledge Gagné et al. 2014 (http://adsabs.harvard.edu/abs/2013arXiv1312.5864G) when using this routine. In addition to this, please acknowledge
;       the krEllipsoidFit authors (Ronn Kling and Jerry Lefever), in addition to the website www.rlkling.com
;
; PROCEDURES USED:
;       KRELLIPSOIDFIT, REVERSE(), DETERM(), ROTMATRIX_TO_EULER(), COORD_ROT
;
; MODIFICATION HISTORY:
;       WRITTEN, Jonathan Gagne, January, 11 2014
;-
Pro fit_3D_ellipsoid, X, Y, Z, CENTER_XYZ, SIGMA_XYZ_P, rotmatrix, euler_angles, XP, YP, ZP, EX=ex, EY=ey, EZ=ez, EXP=exp, EYP=eyp, EZP=ezp
  
  ;Declare subroutines
  forward_function krEllipsoidFit, reverse, determ, rotmatrix_to_euler, coord_rot
  
  if n_params() lt 3. then $
    message, ' You must supply a set of cartesian X, Y, Z coordinates ! Syntax : fit_3D_ellipsoid, X, Y, Z, CENTER_XYZ, SIGMA_XYZ_P, rotmatrix, euler_angles, XP, YP, ZP, EX=ex, EY=ey, EZ=ez, EXP=exp, EYP=eyp, EZP=ezp' 
  
  nobj = n_elements(x)
  if n_elements(y) ne nobj or n_elements(z) ne nobj then $
    message, ' All X, Y, Z arrays must have the same lenght !'
  if nobj lt 7. then $
    message, ' It is not recommended to use this routine with fewer than 7 XYZ coordinates, since the system will be under-determined !'
  
  ;Fit a free-to-rotate ellipsoid in XYZ coordinates
  krEllipsoidFit, X, Y, Z, evec=evec_xyz, XM=center_xyz, evals=sigma_xyz_p_0
  
  ;Compute rotation matrices from eigenvectors output by krEllipsoidFit
  rotmatrix = reverse(transpose(evec_xyz))
  
  ;Compute 1-sigma scatter in each direction using eigenvalues output by krEllipsoidFit
  sigma_xyz_p = reverse(sigma_xyz_p_0)
  
  ;Force the determinant of the rotation matrix to be +1 (krEllipsoidFits sometimes generates a rotation + reflection with determinant -1, which messes up the transformation to Euler angles)
  rotmatrix *= sign(determ(rotmatrix))
  
  ;Compute the (psi, theta, phi) Euler angles 
  euler_angles = rotmatrix_to_euler(rotmatrix,/DEGREE)
  
  ;Set uncertainties on individual cartesian positions to zero if they're not explicitely given
  if ~keyword_set(xp) then xp = fltarr(n_elements(x))
  if ~keyword_set(yp) then yp = fltarr(n_elements(y))
  if ~keyword_set(zp) then zp = fltarr(n_elements(z))
  
  ;Rotate cartesian data points to the new set of primed coordinates (XP, YP, ZP) and propagate errors
  coord_rot, x, y, z, xp, yp, zp, ex, ey, ez, exp, eyp, ezp, ROT_MATRIX=rotmatrix, CENTER=center_xyz
  
End